"""
Constants for swarm-bus
"""

PREFIX = 'dev-%(hostname)s-'
PRIORITIES = ['low', 'medium', 'high']
OFFICE_HOURS = True
POLLING_INTERVAL = 20
RETENTION_PERIOD = 864000  # 10 days
VISIBILITY_TIMEOUT = 30
SLEEP_TIME = 0
REGION = 'eu-west-1'



"""
Connector for swarm-bus
"""
import logging

import boto3

from swarm_bus.transport import Transport

logger = logging.getLogger(__name__)


class BusConnector(object):
    """
    Bus connector logic
    """
    log_namespace = 'Swarm-Bus'
    connection = None
    client = None

    def __init__(self, **transport):
        """
        Handle transport and connection
        """
        self.queues = {}
        self.transport = Transport(**transport)

    def connect(self):
        """
        Connect to SQS resource
        """
        session = boto3.session.Session(
            aws_access_key_id=self.transport.access_key,
            aws_secret_access_key=self.transport.secret_key,
            region_name=self.transport.region
        )

        self.connection = session.resource('sqs', use_ssl=True)
        self.client = self.connection.meta.client

        logger.debug(
            "[%s] Connected to SQS, using '%s' prefix",
            self.log_namespace,
            self.transport.prefix
        )

    def disconnect(self):
        """
        Disconnect to SQS resource
        """
        if self.connection:
            self.queues = {}
            self.client = None
            self.connection = None
            logger.debug(
                '[%s] Disconnected from SQS',
                self.log_namespace
            )


class BusContextManager(object):
    """
    Context Manager features for AMQP.
    """
    def __del__(self):
        """
        Close the connection when garbage collected.
        """
        self.disconnect()

    def __enter__(self):
        """
        Context manager establishing connection.
        """
        if self.connection is None:
            self.connect()
        return self

    def __exit__(self, exception_type, exception_value, traceback):
        """
        Context manager closing connection.
        """
        self.disconnect()



"""
Consumer for swarm-bus
"""
import base64
import datetime
import json
import logging
import time

logger = logging.getLogger(__name__)


class Consumer(object):
    """
    Consumer logic
    """
    polling_messages = 10

    def consume(self, queue_name, callback=None, error_handler=None,
                polling_messages=10, max_messages=100000,
                pre_polling_handler=None, post_polling_handler=None):
        """
        Consume message datas
        """
        self.polling_messages = polling_messages

        queue_set = self.get_queue_set(queue_name)

        if callback is None:
            raise ValueError(
                "callback parameter can not be empty"
            )

        callback_wrapped = self.callback_wrapper(callback, error_handler)

        while max_messages > 0:
            if not self.can_consume:
                logger.debug(
                    '[%s] Consuming is on hold. Next try in 60 seconds',
                    self.log_namespace
                )
                sleep_time = 60
            else:
                if pre_polling_handler:
                    pre_polling_handler(max_messages)

                polled = self.polling(
                    queue_set, callback_wrapped,
                    self.polling_messages
                )

                if post_polling_handler:
                    post_polling_handler(max_messages, polled)

                max_messages -= polled

                sleep_time = queue_set.sleep_time
                if sleep_time:
                    logger.debug(
                        '[%s] Sleeping for %s seconds',
                        self.log_namespace,
                        sleep_time
                    )

            time.sleep(sleep_time)

        logger.debug(
            '[%s] Stop consuming messages',
            self.log_namespace
        )

    @property
    def can_consume(self):
        """
        Check a queue can be consumed.
        """
        if not self.transport.office_hours:
            return True

        now = datetime.datetime.now()
        if now.weekday() in [5, 6]:  # Week-end
            return False

        hour = now.hour
        if hour >= 8 and hour < 20:
            return True

        return False

    def polling(self, queue_set, callback_wrapped, polling_messages):
        """
        Poll all the messages availables for each priorities
        """
        total_polled = 0

        for queue_priority in queue_set.queues:
            logger.debug(
                '[%s] Polling %s for %s seconds',
                self.log_namespace,
                queue_priority.url,
                queue_set.polling_interval
            )
            polled_messages = queue_priority.receive_messages(
                WaitTimeSeconds=queue_set.polling_interval,
                MaxNumberOfMessages=polling_messages
            )
            logger.debug(
                '[%s] %s messages polled',
                self.log_namespace,
                len(polled_messages)
            )
            total_polled += len(polled_messages)

            for message in polled_messages:
                body = self.decode_body(message)

                callback_wrapped(
                    body, message
                )

        return total_polled

    def decode_body(self, message):
        """
        Decode the message body for having a dict
        """
        try:
            return json.loads(message.body)
        except ValueError:
            # Old bus compatibility
            return json.loads(
                base64.b64decode(
                    json.loads(
                        base64.b64decode(
                            message.body
                        ).decode('utf8')
                    )['body']
                ).decode('utf8')
            )

    def callback_wrapper(self, callback, error_handler):
        """
        Decorate the callback to log exceptions
        and send them to Senty later if possible.

        Also cancels the exception to avoid process to crash !
        """
        def exception_catcher(body, message):
            """
            Decorator around callback.
            """
            try:
                return callback(body, message)
            except Exception:
                logger.exception(
                    '[%s] Unhandled exception occured !',
                    self.log_namespace
                )
                if error_handler:
                    error_handler(body, message)
                    logger.debug(
                        '[%s] Error handler called',
                        self.log_namespace
                    )

        return exception_catcher



"""
Swarm Bus
"""
from swarm_bus.connector import BusConnector
from swarm_bus.connector import BusContextManager
from swarm_bus.consumer import Consumer
from swarm_bus.deleter import Deleter
from swarm_bus.lister import Lister
from swarm_bus.publisher import Publisher
from swarm_bus.purger import Purger
from swarm_bus.queue import QueueManager


class SwarmBus(
        BusConnector,
        BusContextManager,
        QueueManager,
        Deleter,
        Lister,
        Purger,
        Publisher,
        Consumer
):
    pass
